/*
 * Copyright 2013-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * This program and the accompanying materials are made available under the 
 * terms of the Eclipse Public License v1.0 which accompanies this distribution,
 * and is available at http://www.eclipse.org/legal/epl-v10.html.
 */

package com.aquenos.csstudio.archive.json.reader;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.math.BigInteger;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.WeakHashMap;
import java.util.zip.DeflaterInputStream;
import java.util.zip.GZIPInputStream;

import org.codehaus.jackson.JsonFactory;
import org.codehaus.jackson.JsonLocation;
import org.codehaus.jackson.JsonParseException;
import org.codehaus.jackson.JsonParser;
import org.codehaus.jackson.JsonParser.Feature;
import org.codehaus.jackson.JsonToken;
import org.csstudio.archive.reader.ArchiveInfo;
import org.csstudio.archive.reader.ArchiveReader;
import org.csstudio.archive.reader.UnknownChannelException;
import org.csstudio.archive.reader.ValueIterator;
import org.epics.util.array.ArrayDouble;
import org.epics.util.array.ArrayInt;
import org.epics.util.array.ArrayLong;
import org.epics.util.text.NumberFormats;
import org.epics.util.time.Timestamp;
import org.epics.vtype.AlarmSeverity;
import org.epics.vtype.Display;
import org.epics.vtype.VType;
import org.epics.vtype.ValueFactory;

import com.aquenos.csstudio.archive.json.reader.internal.DoubleArrayStatistics;
import com.aquenos.csstudio.archive.json.reader.internal.DoubleStatistics;

/**
 * Archive reader implementation that connects to an archive server using an
 * HTTP / JSON based protocol. Typically, this reader is used together with the
 * JSON archive server. However, it will work with any compliant HTTP server.
 * 
 * @author Sebastian Marsching
 */
public class JsonArchiveReader implements ArchiveReader {

    private final static BigInteger ONE_BILLION = BigInteger
            .valueOf(1000000000L);

    private String url;
    private String httpUrl;
    private JsonFactory jsonFactory;
    private ArchiveInfo[] archiveInfos;
    private WeakHashMap<JsonValueIterator, Void> iterators = new WeakHashMap<JsonArchiveReader.JsonValueIterator, Void>();

    /**
     * Creates an archive reader that requests samples from the specified URL.
     * The URL must start with the the scheme "json" followed by the HTTP or
     * HTTPS URL of the archive server. The URL must include the context path,
     * but not include the servlet path. For example the URL
     * "json:http://localhost:8080/" will expect the archive server to run on
     * port 8080 of the same computer and will use the URL
     * "http://localhost:8080/archive/" for requesting a list of the archives
     * available on this server. The information about the available archives
     * will only be loaded once on construction and reused for the whole
     * lifetime of this object.
     * 
     * @param url
     *            archive URL with the scheme "json" followed by a valid HTTP or
     *            HTTPS URL.
     * @throws IOException
     *             if information about archives cannot be loaded from the
     *             archive server.
     */
    public JsonArchiveReader(String url) throws IOException {
        this.url = url;
        if (!this.url.startsWith("json:")) {
            throw new IllegalArgumentException(
                    "The URL \""
                            + this.url
                            + "\" is not a valid archive URL, because it does not start with \"json:\".");
        }
        this.httpUrl = this.url.substring(5);
        if (!this.httpUrl.endsWith("/")) {
            this.httpUrl = httpUrl + "/";
        }
        this.jsonFactory = new JsonFactory();
        jsonFactory.enable(Feature.ALLOW_NON_NUMERIC_NUMBERS);
        loadArchiveInfos();
    }

    @Override
    public String getServerName() {
        return "JSON Archive";
    }

    @Override
    public String getURL() {
        return url;
    }

    @Override
    public String getDescription() {
        return "Provides archive access over HTTP/JSON.";
    }

    @Override
    public int getVersion() {
        return 0;
    }

    private void loadArchiveInfos() throws IOException {
        InputStream is = doGet("/");
        JsonParser parser = jsonFactory.createJsonParser(is);
        JsonToken token = parser.nextToken();
        if (token == null) {
            throw new IOException("Unexpected end of stream.");
        }
        if (token != JsonToken.START_ARRAY) {
            throw new JsonParseException("Expected START_ARRAY but got "
                    + token, parser.getTokenLocation());
        }
        LinkedList<ArchiveInfo> archiveInfos = new LinkedList<ArchiveInfo>();
        while (true) {
            token = parser.nextToken();
            if (token == null) {
                throw new IOException("Unexpected end of stream.");
            }
            if (token == JsonToken.END_ARRAY) {
                break;
            }
            ArchiveInfo archiveInfo = readArchiveInfo(parser);
            archiveInfos.add(archiveInfo);
        }
        parser.close();
        this.archiveInfos = archiveInfos.toArray(new ArchiveInfo[archiveInfos
                .size()]);
    }

    private ArchiveInfo readArchiveInfo(JsonParser parser) throws IOException {
        JsonToken token = parser.getCurrentToken();
        if (token != JsonToken.START_OBJECT) {
            throw new JsonParseException("Expected START_OBJECT but got "
                    + token, parser.getTokenLocation());
        }
        Integer archiveKey = null;
        String archiveName = null;
        String archiveDescription = null;
        String fieldName = null;
        while (true) {
            token = parser.nextToken();
            if (token == null) {
                throw new IOException("Unexpected end of stream.");
            }
            if (token == JsonToken.END_OBJECT) {
                break;
            }
            if (fieldName == null) {
                if (token == JsonToken.FIELD_NAME) {
                    fieldName = parser.getCurrentName();
                } else {
                    throw new JsonParseException("Expected FIELD_NAME but got "
                            + token, parser.getTokenLocation());
                }
            } else {
                if (fieldName.equals("key")) {
                    if (token == JsonToken.VALUE_NUMBER_INT) {
                        if (archiveKey != null) {
                            throw new JsonParseException("Field \"" + fieldName
                                    + "\" occurs twice.",
                                    parser.getTokenLocation());
                        }
                        archiveKey = parser.getIntValue();
                    } else {
                        throw new JsonParseException(
                                "Expected VALUE_NUMBER_INT but got " + token,
                                parser.getTokenLocation());
                    }
                } else if (fieldName.equals("name")) {
                    if (token == JsonToken.VALUE_STRING) {
                        if (archiveName != null) {
                            throw new JsonParseException("Field \"" + fieldName
                                    + "\" occurs twice.",
                                    parser.getTokenLocation());
                        }
                        archiveName = parser.getText();
                    } else {
                        throw new JsonParseException(
                                "Expected VALUE_STRING but got " + token,
                                parser.getTokenLocation());
                    }
                } else if (fieldName.equals("description")) {
                    if (token == JsonToken.VALUE_STRING) {
                        if (archiveDescription != null) {
                            throw new JsonParseException("Field \"" + fieldName
                                    + "\" occurs twice.",
                                    parser.getTokenLocation());
                        }
                        archiveDescription = parser.getText();
                    } else {
                        throw new JsonParseException(
                                "Expected VALUE_STRING but got " + token,
                                parser.getTokenLocation());
                    }
                } else {
                    throw new JsonParseException("Found unknown field \""
                            + fieldName + "\".", parser.getTokenLocation());
                }
                fieldName = null;
            }
        }
        if (archiveKey == null || archiveName == null
                || archiveDescription == null) {
            throw new JsonParseException(
                    "Mandatory field is missing in object.",
                    parser.getTokenLocation());
        }
        return new ArchiveInfo(archiveName, archiveDescription, archiveKey);
    }

    @Override
    public ArchiveInfo[] getArchiveInfos() {
        // Arrays cannot be made read-only, thus we return a copy.
        return Arrays.copyOf(archiveInfos, archiveInfos.length);
    }

    @Override
    public String[] getNamesByPattern(int key, String globPattern)
            throws Exception {
        return getNames("/" + key + "/channels-by-pattern/"
                + URLEncoder.encode(globPattern, "UTF-8"));
    }

    @Override
    public String[] getNamesByRegExp(int key, String regExp) throws Exception {
        return getNames("/" + key + "/channels-by-regexp/"
                + URLEncoder.encode(regExp, "UTF-8"));
    }

    private String[] getNames(String url) throws IOException {
        InputStream is = doGet(url);
        JsonParser parser = jsonFactory.createJsonParser(is);
        JsonToken token = parser.nextToken();
        if (token == null) {
            throw new IOException("Unexpected end of stream.");
        }
        if (token != JsonToken.START_ARRAY) {
            throw new JsonParseException("Expected START_ARRAY but got "
                    + token, parser.getTokenLocation());
        }
        LinkedList<String> channelNames = new LinkedList<String>();
        while (true) {
            token = parser.nextToken();
            if (token == null) {
                throw new IOException("Unexpected end of stream.");
            }
            if (token == JsonToken.END_ARRAY) {
                break;
            }
            if (token == JsonToken.VALUE_STRING) {
                String channelName = parser.getText();
                channelNames.add(channelName);
            } else {
                throw new JsonParseException("Expected VALUE_STRING but got "
                        + token, parser.getTokenLocation());
            }
        }
        parser.close();
        return channelNames.toArray(new String[channelNames.size()]);
    }

    @Override
    public ValueIterator getRawValues(int key, String name, Timestamp start,
            Timestamp end) throws UnknownChannelException, Exception {
        return new JsonValueIterator(key, name, start, end, null);
    }

    @Override
    public ValueIterator getOptimizedValues(int key, String name,
            Timestamp start, Timestamp end, int count)
            throws UnknownChannelException, Exception {
        return new JsonValueIterator(key, name, start, end, count);
    }

    private BigInteger timestampToBigInteger(Timestamp timestamp) {
        return BigInteger.valueOf(timestamp.getNanoSec()).add(
                BigInteger.valueOf(timestamp.getSec()).multiply(ONE_BILLION));
    }

    private Timestamp bigIntegerToTimestamp(BigInteger bigInt) {
        BigInteger[] quotientAndRemainder = bigInt
                .divideAndRemainder(ONE_BILLION);
        return Timestamp.of(quotientAndRemainder[0].longValue(),
                (int) quotientAndRemainder[1].longValue());
    }

    @Override
    public void cancel() {
        synchronized (iterators) {
            for (JsonValueIterator i : iterators.keySet()) {
                i.cancel();
            }
        }
    }

    @Override
    public void close() {
        // We do nothing here, because we do not hold any expensive resources
        // that need to be closed.
    }

    private InputStream doGet(String url) throws IOException {
        String requestUrl = this.httpUrl + "archive" + url;
        URLConnection connection = new URL(requestUrl).openConnection();
        connection.addRequestProperty("Accept-Encoding", "gzip, deflate");
        connection.connect();
        String contentEncoding = connection.getHeaderField("Content-Encoding");
        if (contentEncoding != null) {
            if (contentEncoding.equals("gzip")) {
                return new GZIPInputStream(connection.getInputStream());
            } else if (contentEncoding.equals("deflate")) {
                return new DeflaterInputStream(connection.getInputStream());
            }
        }
        return connection.getInputStream();
    }

    private static enum Quality {
        ORIGINAL, INTERPOLATED
    }

    private class JsonValueIterator implements ValueIterator {

        private JsonParser parser;
        private VType nextValue;
        private boolean canceled = false;
        private final Object cancelLock = new Object();
        private boolean finished = false;

        public JsonValueIterator(int key, String name, Timestamp start,
                Timestamp end, Integer count) throws IOException,
                UnknownChannelException {
            StringBuilder sb = new StringBuilder();
            sb.append("/");
            sb.append(key);
            sb.append("/samples/");
            sb.append(URLEncoder.encode(name, "UTF-8"));
            sb.append("?start=");
            sb.append(timestampToBigInteger(start).toString());
            sb.append("&end=");
            sb.append(timestampToBigInteger(end).toString());
            if (count != null) {
                sb.append("&count=");
                sb.append(count);
            }
            try {
                InputStream is = doGet(sb.toString());
                this.parser = jsonFactory.createJsonParser(is);
                JsonToken token = parser.nextToken();
                if (token == null) {
                    throw new IOException("Unexpected end of stream.");
                }
                if (token != JsonToken.START_ARRAY) {
                    throw new JsonParseException(
                            "Expected START_ARRAY but got " + token,
                            parser.getTokenLocation());
                }
            } catch (FileNotFoundException e) {
                throw new UnknownChannelException(name);
            }
            synchronized (JsonArchiveReader.this.iterators) {
                JsonArchiveReader.this.iterators.put(this, null);
            }
        }

        @Override
        public boolean hasNext() {
            try {
                return hasNextInternal();
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }

        private boolean hasNextInternal() throws IOException {
            if (nextValue != null) {
                return true;
            }
            if (finished) {
                return false;
            }
            boolean canceled;
            synchronized (cancelLock) {
                canceled = this.canceled;
            }
            if (canceled) {
                if (parser != null) {
                    try {
                        parser.close();
                    } catch (IOException e) {
                        // Ignore exception on clean-up.
                    }
                    parser = null;
                }
                throw new IllegalStateException(
                        "The operation has been canceled.");
            }
            if (parser == null) {
                return false;
            }
            JsonToken token = parser.nextToken();
            if (token == null) {
                throw new IOException("Unexpected end of stream.");
            }
            if (token == JsonToken.END_ARRAY) {
                try {
                    parser.close();
                } catch (IOException e) {
                    // Ignore exception on clean-up.
                }
                parser = null;
                finished = true;
                return false;
            }
            nextValue = readValue(parser);
            return true;
        }

        private VType readValue(JsonParser parser) throws IOException {
            JsonToken token = parser.getCurrentToken();
            if (token != JsonToken.START_OBJECT) {
                throw new JsonParseException("Expected START_OBJECT but got "
                        + token, parser.getTokenLocation());
            }
            String fieldName = null;
            String type = null;
            Timestamp time = null;
            AlarmSeverity severity = null;
            String status = null;
            Quality quality = null;
            Display display = null;
            String[] labels = null;
            Object value = null;
            Double minimum = null;
            Double maximum = null;
            while (true) {
                token = parser.nextToken();
                if (token == null) {
                    throw new IOException("Unexpected end of stream.");
                }
                if (token == JsonToken.END_OBJECT) {
                    break;
                }
                if (fieldName == null) {
                    if (token != JsonToken.FIELD_NAME) {
                        throw new JsonParseException(
                                "Expected FIELD_NAME but got " + token,
                                parser.getTokenLocation());
                    }
                    fieldName = parser.getCurrentName();
                } else {
                    if (fieldName.equals("time")) {
                        if (time != null) {
                            throw new JsonParseException("Field \"" + fieldName
                                    + "\" occurs twice.",
                                    parser.getTokenLocation());
                        }
                        if (token != JsonToken.VALUE_NUMBER_INT) {
                            throw new JsonParseException(
                                    "Expected VALUE_NUMBER_INT but got "
                                            + token, parser.getTokenLocation());
                        }
                        time = bigIntegerToTimestamp(parser
                                .getBigIntegerValue());
                    } else if (fieldName.equals("severity")) {
                        if (severity != null) {
                            throw new JsonParseException("Field \"" + fieldName
                                    + "\" occurs twice.",
                                    parser.getTokenLocation());
                        }
                        severity = readSeverity(parser);
                    } else if (fieldName.equals("status")) {
                        if (status != null) {
                            throw new JsonParseException("Field \"" + fieldName
                                    + "\" occurs twice.",
                                    parser.getTokenLocation());
                        }
                        if (token != JsonToken.VALUE_STRING) {
                            throw new JsonParseException(
                                    "Expected VALUE_STRING but got " + token,
                                    parser.getTokenLocation());
                        }
                        status = parser.getText();
                    } else if (fieldName.equals("quality")) {
                        if (quality != null) {
                            throw new JsonParseException("Field \"" + fieldName
                                    + "\" occurs twice.",
                                    parser.getTokenLocation());
                        }
                        quality = readQuality(parser);
                    } else if (fieldName.equals("metaData")) {
                        if (labels != null || display != null) {
                            throw new JsonParseException("Field \"" + fieldName
                                    + "\" occurs twice.",
                                    parser.getTokenLocation());
                        }
                        Object metaData = readMetaData(parser);
                        if (metaData instanceof Display) {
                            display = (Display) metaData;
                        } else if (metaData instanceof String[]) {
                            labels = (String[]) metaData;
                        } else {
                            throw new RuntimeException(
                                    "Return value of internal method readMetaData had unexpected type "
                                            + metaData.getClass().getName()
                                            + ".");
                        }
                    } else if (fieldName.equals("type")) {
                        if (type != null) {
                            throw new JsonParseException("Field \"" + fieldName
                                    + "\" occurs twice.",
                                    parser.getTokenLocation());
                        }
                        if (token != JsonToken.VALUE_STRING) {
                            throw new JsonParseException(
                                    "Expected VALUE_STRING but got " + token,
                                    parser.getTokenLocation());
                        }
                        type = parser.getText();
                    } else if (fieldName.equals("value")) {
                        if (value != null) {
                            throw new JsonParseException("Field \"" + fieldName
                                    + "\" occurs twice.",
                                    parser.getTokenLocation());
                        }
                        if (type == null) {
                            throw new JsonParseException(
                                    "\"value\" field must be specified after \"type\" field.",
                                    parser.getTokenLocation());
                        }
                        if (type.equalsIgnoreCase("double")) {
                            value = readDoubleArray(parser);
                        } else if (type.equalsIgnoreCase("minMaxDouble")) {
                            value = readDoubleArray(parser);
                        } else if (type.equalsIgnoreCase("long")) {
                            value = readLongArray(parser);
                        } else if (type.equalsIgnoreCase("enum")) {
                            value = readIntArray(parser);
                        } else if (type.equalsIgnoreCase("string")) {
                            value = readStringArray(parser);
                        } else {
                            throw new JsonParseException(
                                    "Cannot parse value for unknown type \""
                                            + type + "\".",
                                    parser.getTokenLocation());
                        }
                    } else if (fieldName.equals("minimum")) {
                        if (minimum != null) {
                            throw new JsonParseException("Field \"" + fieldName
                                    + "\" occurs twice.",
                                    parser.getTokenLocation());
                        }
                        if (token != JsonToken.VALUE_NUMBER_INT
                                && token != JsonToken.VALUE_NUMBER_FLOAT) {
                            if (token != JsonToken.VALUE_STRING) {
                                throw new JsonParseException(
                                        "Expected VALUE_NUMBER_INT, VALUE_NUMBER_FLOAT, or VALUE_STRING but got "
                                                + token,
                                        parser.getTokenLocation());
                            }
                            minimum = stringToSpecialDouble(parser.getText(),
                                    parser.getTokenLocation());
                        } else {
                            minimum = parser.getDoubleValue();
                        }
                    } else if (fieldName.equals("maximum")) {
                        if (maximum != null) {
                            throw new JsonParseException("Field \"" + fieldName
                                    + "\" occurs twice.",
                                    parser.getTokenLocation());
                        }
                        if (token != JsonToken.VALUE_NUMBER_INT
                                && token != JsonToken.VALUE_NUMBER_FLOAT) {
                            if (token != JsonToken.VALUE_STRING) {
                                throw new JsonParseException(
                                        "Expected VALUE_NUMBER_INT, VALUE_NUMBER_FLOAT, or VALUE_STRING but got "
                                                + token,
                                        parser.getTokenLocation());
                            }
                            maximum = stringToSpecialDouble(parser.getText(),
                                    parser.getTokenLocation());
                        } else {
                            maximum = parser.getDoubleValue();
                        }
                    } else {
                        throw new JsonParseException("Found unknown field \""
                                + fieldName + "\".", parser.getTokenLocation());
                    }
                    fieldName = null;
                }
            }
            if (type == null || time == null || severity == null
                    || status == null || quality == null || value == null) {
                throw new JsonParseException(
                        "Mandatory field is missing in object.",
                        parser.getTokenLocation());
            }
            if (type.equalsIgnoreCase("minMaxDouble")) {
                if (labels != null) {
                    throw new JsonParseException(
                            "Value of type \"minMaxDouble\" only accepts numeric meta-data.",
                            parser.getTokenLocation());
                }
                if (minimum == null || maximum == null) {
                    throw new JsonParseException(
                            "Mandatory field is missing in object.",
                            parser.getTokenLocation());
                }
                double[] valueAsDouble = (double[]) value;
                if (valueAsDouble.length == 1) {
                    return new DoubleStatistics(valueAsDouble[0], status,
                            severity, display, minimum, maximum, 0, Double.NaN,
                            null, time, true);
                } else {
                    return new DoubleArrayStatistics(new ArrayDouble(
                            valueAsDouble, true), status, severity, display,
                            minimum, maximum, 0, Double.NaN, null, time, true);
                }
            }
            if (minimum != null || maximum != null) {
                throw new JsonParseException(
                        "Invalid field specifed for value of type\"" + type
                                + "\".", parser.getTokenLocation());
            }
            if (type.equalsIgnoreCase("double")) {
                if (labels != null) {
                    throw new JsonParseException(
                            "Value of type \"double\" only accepts numeric meta-data.",
                            parser.getTokenLocation());
                }
                double[] valueAsDouble = (double[]) value;
                if (valueAsDouble.length == 1) {
                    return ValueFactory.newVNumber(valueAsDouble[0],
                            ValueFactory.newAlarm(severity, status),
                            ValueFactory.newTime(time), display);
                } else {
                    return ValueFactory.newVNumberArray(new ArrayDouble(
                            valueAsDouble, true), ValueFactory.newAlarm(
                            severity, status), ValueFactory.newTime(time),
                            display);
                }
            }
            if (type.equalsIgnoreCase("long")) {
                if (labels != null) {
                    throw new JsonParseException(
                            "Value of type \"long\" only accepts numeric meta-data.",
                            parser.getTokenLocation());
                }
                long[] valueAsLong = (long[]) value;
                if (valueAsLong.length == 1) {
                    return ValueFactory.newVLong(valueAsLong[0],
                            ValueFactory.newAlarm(severity, status),
                            ValueFactory.newTime(time), display);
                } else {
                    return ValueFactory.newVLongArray(new ArrayLong(
                            valueAsLong, true), ValueFactory.newAlarm(severity,
                            status), ValueFactory.newTime(time), display);
                }
            }
            if (type.equalsIgnoreCase("enum")) {
                if (display != null) {
                    throw new JsonParseException(
                            "Value of type \"enum\" only accepts enum meta-data.",
                            parser.getTokenLocation());
                }
                List<String> labelsList;
                if (labels != null) {
                    labelsList = Arrays.asList(labels);
                } else {
                    labelsList = Collections.emptyList();
                }
                int[] valueAsInt = (int[]) value;
                if (valueAsInt.length == 1) {
                    return ValueFactory.newVEnum(valueAsInt[0], labelsList,
                            ValueFactory.newAlarm(severity, status),
                            ValueFactory.newTime(time));
                } else {
                    return ValueFactory.newVEnumArray(new ArrayInt(valueAsInt,
                            true), labelsList, ValueFactory.newAlarm(severity,
                            status), ValueFactory.newTime(time));
                }
            }
            if (type.equalsIgnoreCase("string")) {
                if (labels != null || display != null) {
                    throw new JsonParseException(
                            "Value of type \"string\" does not accept meta-data.",
                            parser.getTokenLocation());
                }
                String[] valueAsString = (String[]) value;
                if (valueAsString.length == 1) {
                    return ValueFactory.newVString(valueAsString[0],
                            ValueFactory.newAlarm(severity, status),
                            ValueFactory.newTime(time));
                } else {
                    return ValueFactory.newVStringArray(
                            Arrays.asList(valueAsString),
                            ValueFactory.newAlarm(severity, status),
                            ValueFactory.newTime(time));
                }
            }
            throw new JsonParseException(
                    "Invalid value type \"" + type + "\".",
                    parser.getTokenLocation());
        }

        private AlarmSeverity readSeverity(JsonParser parser)
                throws IOException {
            JsonToken token = parser.getCurrentToken();
            if (token != JsonToken.START_OBJECT) {
                throw new JsonParseException("Expected START_OBJECT but got "
                        + token, parser.getTokenLocation());
            }
            String fieldName = null;
            String levelString = null;
            Boolean hasValue = null;
            while (true) {
                token = parser.nextToken();
                if (token == null) {
                    throw new IOException("Unexpected end of stream.");
                }
                if (token == JsonToken.END_OBJECT) {
                    break;
                }
                if (fieldName == null) {
                    if (token != JsonToken.FIELD_NAME) {
                        throw new JsonParseException(
                                "Expected FIELD_NAME but got " + token,
                                parser.getTokenLocation());
                    }
                    fieldName = parser.getCurrentName();
                } else {
                    if (fieldName.equals("level")) {
                        if (levelString != null) {
                            throw new JsonParseException("Field \"" + fieldName
                                    + "\" occurs twice.",
                                    parser.getTokenLocation());
                        }
                        if (token != JsonToken.VALUE_STRING) {
                            throw new JsonParseException(
                                    "Expected VALUE_STRING but got " + token,
                                    parser.getTokenLocation());
                        }
                        levelString = parser.getText();
                    } else if (fieldName.equals("hasValue")) {
                        if (hasValue != null) {
                            throw new JsonParseException("Field \"" + fieldName
                                    + "\" occurs twice.",
                                    parser.getTokenLocation());
                        }
                        if (token != JsonToken.VALUE_TRUE
                                && token != JsonToken.VALUE_FALSE) {
                            throw new JsonParseException(
                                    "Expected VALUE_TRUE or VALUE_FALSE but got "
                                            + token, parser.getTokenLocation());
                        }
                        hasValue = parser.getBooleanValue();
                    } else {
                        throw new JsonParseException("Found unknown field \""
                                + fieldName + "\".", parser.getTokenLocation());
                    }
                    fieldName = null;
                }
            }
            if (hasValue == null || levelString == null) {
                throw new JsonParseException(
                        "Mandatory field is missing in object.",
                        parser.getTokenLocation());
            }
            final AlarmSeverity severity;
            if (levelString.equalsIgnoreCase("OK")) {
                severity = AlarmSeverity.NONE;
            } else if (levelString.equalsIgnoreCase("MINOR")) {
                severity = AlarmSeverity.MINOR;
            } else if (levelString.equalsIgnoreCase("MAJOR")) {
                severity = AlarmSeverity.MAJOR;
            } else if (levelString.equalsIgnoreCase("INVALID")) {
                severity = AlarmSeverity.INVALID;
            } else {
                throw new JsonParseException("Unknown severity \""
                        + levelString + "\".", parser.getTokenLocation());
            }
            return severity;
        }

        private Quality readQuality(JsonParser parser) throws IOException {
            JsonToken token = parser.getCurrentToken();
            if (token != JsonToken.VALUE_STRING) {
                throw new JsonParseException("Expected VALUE_STRING but got "
                        + token, parser.getTokenLocation());
            }
            String qualityString = parser.getText();
            if (qualityString.equalsIgnoreCase("Original")) {
                return Quality.ORIGINAL;
            } else if (qualityString.equalsIgnoreCase("Interpolated")) {
                return Quality.INTERPOLATED;
            } else {
                throw new JsonParseException("Unknown quality value: \""
                        + qualityString + "\"", parser.getTokenLocation());
            }
        }

        /**
         * Reads the meta-data associated with a value. There are different
         * types of meta-data for numeric and enum values, therefore the type of
         * the return value has to be determined at runtime.
         * 
         * @param parser
         *            the JSON parser that is used to read the meta-data.
         * @return a instance of {@link String}<code>[]</code> (storing the enum
         *         labels) or an instance of {@link Display} (storing numeric
         *         limits and number formatting information).
         * @throws IOException
         *             if an error occurs while parsing the JSON input (e.g.
         *             interrupted stream, malformed data).
         */
        private Object readMetaData(JsonParser parser) throws IOException {
            JsonToken token = parser.getCurrentToken();
            if (token == null) {
                throw new IOException("Unexpected end of stream.");
            }
            if (token != JsonToken.START_OBJECT) {
                throw new JsonParseException("Expected START_OBJECT but got "
                        + token, parser.getTokenLocation());
            }
            String fieldName = null;
            String type = null;
            Integer precision = null;
            String units = null;
            Double displayLow = null;
            Double displayHigh = null;
            Double warnLow = null;
            Double warnHigh = null;
            Double alarmLow = null;
            Double alarmHigh = null;
            String[] states = null;
            while (true) {
                token = parser.nextToken();
                if (token == null) {
                    throw new IOException("Unexpected end of stream.");
                }
                if (token == JsonToken.END_OBJECT) {
                    break;
                }
                if (fieldName == null) {
                    if (token != JsonToken.FIELD_NAME) {
                        throw new JsonParseException(
                                "Expected FIELD_NAME but got " + token,
                                parser.getTokenLocation());
                    }
                    fieldName = parser.getCurrentName();
                } else {
                    if (fieldName.equals("type")) {
                        if (type != null) {
                            throw new JsonParseException("Field \"" + fieldName
                                    + "\" occurs twice.",
                                    parser.getTokenLocation());
                        }
                        if (token != JsonToken.VALUE_STRING) {
                            throw new JsonParseException(
                                    "Expected VALUE_STRING but got " + token,
                                    parser.getTokenLocation());
                        }
                        type = parser.getText();
                    } else if (fieldName.equals("precision")) {
                        if (precision != null) {
                            throw new JsonParseException("Field \"" + fieldName
                                    + "\" occurs twice.",
                                    parser.getTokenLocation());
                        }
                        if (token != JsonToken.VALUE_NUMBER_INT) {
                            throw new JsonParseException(
                                    "Expected VALUE_NUMBER_INT but got "
                                            + token, parser.getTokenLocation());
                        }
                        precision = parser.getIntValue();
                    } else if (fieldName.equals("units")) {
                        if (units != null) {
                            throw new JsonParseException("Field \"" + fieldName
                                    + "\" occurs twice.",
                                    parser.getTokenLocation());
                        }
                        if (token != JsonToken.VALUE_STRING) {
                            throw new JsonParseException(
                                    "Expected VALUE_STRING but got " + token,
                                    parser.getTokenLocation());
                        }
                        units = parser.getText();
                    } else if (fieldName.equals("displayLow")) {
                        if (displayLow != null) {
                            throw new JsonParseException("Field \"" + fieldName
                                    + "\" occurs twice.",
                                    parser.getTokenLocation());
                        }
                        if (token != JsonToken.VALUE_NUMBER_INT
                                && token != JsonToken.VALUE_NUMBER_FLOAT) {
                            if (token != JsonToken.VALUE_STRING) {
                                throw new JsonParseException(
                                        "Expected VALUE_NUMBER_INT, VALUE_NUMBER_FLOAT, or VALUE_STRING but got "
                                                + token,
                                        parser.getTokenLocation());
                            }
                            displayLow = stringToSpecialDouble(
                                    parser.getText(), parser.getTokenLocation());
                        } else {
                            displayLow = parser.getDoubleValue();
                        }
                    } else if (fieldName.equals("displayHigh")) {
                        if (displayHigh != null) {
                            throw new JsonParseException("Field \"" + fieldName
                                    + "\" occurs twice.",
                                    parser.getTokenLocation());
                        }
                        if (token != JsonToken.VALUE_NUMBER_INT
                                && token != JsonToken.VALUE_NUMBER_FLOAT) {
                            if (token != JsonToken.VALUE_STRING) {
                                throw new JsonParseException(
                                        "Expected VALUE_NUMBER_INT, VALUE_NUMBER_FLOAT, or VALUE_STRING but got "
                                                + token,
                                        parser.getTokenLocation());
                            }
                            displayHigh = stringToSpecialDouble(
                                    parser.getText(), parser.getTokenLocation());
                        } else {
                            displayHigh = parser.getDoubleValue();
                        }
                    } else if (fieldName.equals("warnLow")) {
                        if (warnLow != null) {
                            throw new JsonParseException("Field \"" + fieldName
                                    + "\" occurs twice.",
                                    parser.getTokenLocation());
                        }
                        if (token != JsonToken.VALUE_NUMBER_INT
                                && token != JsonToken.VALUE_NUMBER_FLOAT) {
                            if (token != JsonToken.VALUE_STRING) {
                                throw new JsonParseException(
                                        "Expected VALUE_NUMBER_INT, VALUE_NUMBER_FLOAT, or VALUE_STRING but got "
                                                + token,
                                        parser.getTokenLocation());
                            }
                            warnLow = stringToSpecialDouble(parser.getText(),
                                    parser.getTokenLocation());
                        } else {
                            warnLow = parser.getDoubleValue();
                        }
                    } else if (fieldName.equals("warnHigh")) {
                        if (warnHigh != null) {
                            throw new JsonParseException("Field \"" + fieldName
                                    + "\" occurs twice.",
                                    parser.getTokenLocation());
                        }
                        if (token != JsonToken.VALUE_NUMBER_INT
                                && token != JsonToken.VALUE_NUMBER_FLOAT) {
                            if (token != JsonToken.VALUE_STRING) {
                                throw new JsonParseException(
                                        "Expected VALUE_NUMBER_INT, VALUE_NUMBER_FLOAT, or VALUE_STRING but got "
                                                + token,
                                        parser.getTokenLocation());
                            }
                            warnHigh = stringToSpecialDouble(parser.getText(),
                                    parser.getTokenLocation());
                        } else {
                            warnHigh = parser.getDoubleValue();
                        }
                    } else if (fieldName.equals("alarmLow")) {
                        if (alarmLow != null) {
                            throw new JsonParseException("Field \"" + fieldName
                                    + "\" occurs twice.",
                                    parser.getTokenLocation());
                        }
                        if (token != JsonToken.VALUE_NUMBER_INT
                                && token != JsonToken.VALUE_NUMBER_FLOAT) {
                            if (token != JsonToken.VALUE_STRING) {
                                throw new JsonParseException(
                                        "Expected VALUE_NUMBER_INT, VALUE_NUMBER_FLOAT, or VALUE_STRING but got "
                                                + token,
                                        parser.getTokenLocation());
                            }
                            alarmLow = stringToSpecialDouble(parser.getText(),
                                    parser.getTokenLocation());
                        } else {
                            alarmLow = parser.getDoubleValue();
                        }
                    } else if (fieldName.equals("alarmHigh")) {
                        if (alarmHigh != null) {
                            throw new JsonParseException("Field \"" + fieldName
                                    + "\" occurs twice.",
                                    parser.getTokenLocation());
                        }
                        if (token != JsonToken.VALUE_NUMBER_INT
                                && token != JsonToken.VALUE_NUMBER_FLOAT) {
                            if (token != JsonToken.VALUE_STRING) {
                                throw new JsonParseException(
                                        "Expected VALUE_NUMBER_INT, VALUE_NUMBER_FLOAT, or VALUE_STRING but got "
                                                + token,
                                        parser.getTokenLocation());
                            }
                            alarmHigh = stringToSpecialDouble(parser.getText(),
                                    parser.getTokenLocation());
                        } else {
                            alarmHigh = parser.getDoubleValue();
                        }
                    } else if (fieldName.equals("states")) {
                        if (states != null) {
                            throw new JsonParseException("Field \"" + fieldName
                                    + "\" occurs twice.",
                                    parser.getTokenLocation());
                        }
                        states = readStringArray(parser);
                    } else {
                        throw new JsonParseException("Found unknown field \""
                                + fieldName + "\".", parser.getTokenLocation());
                    }
                    fieldName = null;
                }
            }
            if (type == null) {
                throw new JsonParseException(
                        "Mandatory field is missing in object.",
                        parser.getTokenLocation());

            }
            if (type.equalsIgnoreCase("enum")) {
                if (states == null) {
                    throw new JsonParseException(
                            "Mandatory field is missing in object.",
                            parser.getTokenLocation());
                }
                if (precision != null || units != null || displayLow != null
                        || displayHigh != null || warnLow != null
                        || warnHigh != null || alarmLow != null
                        || alarmHigh != null) {
                    throw new JsonParseException(
                            "Invalid field specifed for enum meta-data.",
                            parser.getTokenLocation());
                }
                return states;
            } else if (type.equalsIgnoreCase("numeric")) {
                if (precision == null || units == null || displayLow == null
                        || displayHigh == null || warnLow == null
                        || warnHigh == null || alarmLow == null
                        || alarmHigh == null) {
                    throw new JsonParseException(
                            "Mandatory field is missing in object.",
                            parser.getTokenLocation());
                }
                if (states != null) {
                    throw new JsonParseException(
                            "Invalid field specifed for numeric meta-data.",
                            parser.getTokenLocation());
                }
                return ValueFactory.newDisplay(displayLow, alarmLow, warnLow,
                        units, NumberFormats.format(0), warnHigh, alarmHigh,
                        displayHigh, null, null);
            } else {
                throw new JsonParseException("Invalid meta-data type \"" + type
                        + "\".", parser.getTokenLocation());
            }
        }

        private double[] readDoubleArray(JsonParser parser) throws IOException {
            double[] elements = new double[1];
            int i = 0;
            JsonToken token = parser.getCurrentToken();
            if (token != JsonToken.START_ARRAY) {
                throw new JsonParseException("Expected START_ARRAY but got "
                        + token, parser.getTokenLocation());
            }
            while (true) {
                token = parser.nextToken();
                if (token == null) {
                    throw new IOException("Unexpected end of stream.");
                }
                if (token == JsonToken.END_ARRAY) {
                    break;
                }
                double number;
                if (token == JsonToken.VALUE_NUMBER_INT
                        || token == JsonToken.VALUE_NUMBER_FLOAT) {
                    number = parser.getDoubleValue();
                } else if (token == JsonToken.VALUE_STRING) {
                    number = stringToSpecialDouble(parser.getText(),
                            parser.getTokenLocation());
                } else {
                    throw new JsonParseException(
                            "Expected VALUE_NUMBER_INT, VALUE_NUMBER_FLOAT, or VALUE_STRING but got "
                                    + token, parser.getTokenLocation());

                }
                if (elements.length <= i) {
                    int newLength;
                    if (elements.length <= 1) {
                        newLength = 16;
                    } else {
                        newLength = elements.length * 2;
                    }
                    elements = Arrays.copyOf(elements, newLength);
                }
                elements[i] = number;
                i++;
            }
            if (i == elements.length) {
                return elements;
            } else {
                return Arrays.copyOf(elements, i);
            }
        }

        private long[] readLongArray(JsonParser parser) throws IOException {
            long[] elements = new long[1];
            int i = 0;
            JsonToken token = parser.getCurrentToken();
            if (token != JsonToken.START_ARRAY) {
                throw new JsonParseException("Expected START_ARRAY but got "
                        + token, parser.getTokenLocation());
            }
            while (true) {
                token = parser.nextToken();
                if (token == null) {
                    throw new IOException("Unexpected end of stream.");
                }
                if (token == JsonToken.END_ARRAY) {
                    break;
                }
                if (token == JsonToken.VALUE_NUMBER_INT) {
                    if (elements.length <= i) {
                        int newLength;
                        if (elements.length <= 1) {
                            newLength = 16;
                        } else {
                            newLength = elements.length * 2;
                        }
                        elements = Arrays.copyOf(elements, newLength);
                    }
                    elements[i] = parser.getLongValue();
                    i++;
                } else {
                    throw new JsonParseException(
                            "Expected VALUE_STRING but got " + token,
                            parser.getTokenLocation());
                }
            }
            if (i == elements.length) {
                return elements;
            } else {
                return Arrays.copyOf(elements, i);
            }
        }

        private int[] readIntArray(JsonParser parser) throws IOException {
            int[] elements = new int[1];
            int i = 0;
            JsonToken token = parser.getCurrentToken();
            if (token != JsonToken.START_ARRAY) {
                throw new JsonParseException("Expected START_ARRAY but got "
                        + token, parser.getTokenLocation());
            }
            while (true) {
                token = parser.nextToken();
                if (token == null) {
                    throw new IOException("Unexpected end of stream.");
                }
                if (token == JsonToken.END_ARRAY) {
                    break;
                }
                if (token == JsonToken.VALUE_NUMBER_INT) {
                    if (elements.length <= i) {
                        int newLength;
                        if (elements.length <= 1) {
                            newLength = 16;
                        } else {
                            newLength = elements.length * 2;
                        }
                        elements = Arrays.copyOf(elements, newLength);
                    }
                    elements[i] = parser.getIntValue();
                    i++;
                } else {
                    throw new JsonParseException(
                            "Expected VALUE_STRING but got " + token,
                            parser.getTokenLocation());
                }
            }
            if (i == elements.length) {
                return elements;
            } else {
                return Arrays.copyOf(elements, i);
            }
        }

        private String[] readStringArray(JsonParser parser) throws IOException {
            LinkedList<String> elements = new LinkedList<String>();
            JsonToken token = parser.getCurrentToken();
            if (token != JsonToken.START_ARRAY) {
                throw new JsonParseException("Expected START_ARRAY but got "
                        + token, parser.getTokenLocation());
            }
            while (true) {
                token = parser.nextToken();
                if (token == null) {
                    throw new IOException("Unexpected end of stream.");
                }
                if (token == JsonToken.END_ARRAY) {
                    break;
                }
                if (token == JsonToken.VALUE_STRING) {
                    elements.add(parser.getText());
                } else {
                    throw new JsonParseException(
                            "Expected VALUE_STRING but got " + token,
                            parser.getTokenLocation());
                }
            }
            return elements.toArray(new String[elements.size()]);
        }

        private double stringToSpecialDouble(String value, JsonLocation location)
                throws IOException {
            value = value.toLowerCase();
            if (value.equals("inf") || value.equals("infinity")
                    || value.equals("+inf") || value.equals("+infinity")) {
                return Double.POSITIVE_INFINITY;
            }
            if (value.equals("-inf") || value.equals("-infinity")) {
                return Double.NEGATIVE_INFINITY;
            }
            if (value.equals("nan")) {
                return Double.NaN;
            }
            throw new JsonParseException("String \"" + value
                    + "\" does not qualify as a special double number.",
                    location);
        }

        @Override
        public VType next() throws Exception {
            if (!hasNextInternal()) {
                throw new IllegalStateException("No element is available.");
            }
            VType returnValue = nextValue;
            nextValue = null;
            return returnValue;
        }

        @Override
        public void close() {
            if (parser != null) {
                try {
                    parser.close();
                } catch (IOException e) {
                    // Ignore exception on clean-up.
                }
                parser = null;
                finished = true;
                nextValue = null;
            }
            // Unregister from list of iterators.
            synchronized (JsonArchiveReader.this.iterators) {
                JsonArchiveReader.this.iterators.remove(this);
            }
        }

        public void cancel() {
            synchronized (cancelLock) {
                this.canceled = true;
            }
        }

        @Override
        protected void finalize() throws Throwable {
            this.close();
            super.finalize();
        }

    }

}
